#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017 Google
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# ----------------------------------------------------------------------------
#
#     ***     AUTO GENERATED CODE    ***    AUTO GENERATED CODE     ***
#
# ----------------------------------------------------------------------------
#
#     This file is automatically generated by Magic Modules and manual
#     changes will be clobbered when the file is regenerated.
#
#     Please read more about how to change this file at
#     https://www.github.com/GoogleCloudPlatform/magic-modules
#
# ----------------------------------------------------------------------------

from __future__ import absolute_import, division, print_function

__metaclass__ = type

################################################################################
# Documentation
################################################################################

ANSIBLE_METADATA = {'metadata_version': '1.1', 'status': ["preview"], 'supported_by': 'community'}

DOCUMENTATION = '''
---
module: gcp_compute_instance
description:
- An instance is a virtual machine (VM) hosted on Google's infrastructure.
short_description: Creates a GCP Instance
version_added: 2.6
author: Google Inc. (@googlecloudplatform)
requirements:
- python >= 2.6
- requests >= 2.18.4
- google-auth >= 1.3.0
options:
  state:
    description:
    - Whether the given object should exist in GCP
    choices:
    - present
    - absent
    default: present
    type: str
  can_ip_forward:
    description:
    - Allows this instance to send and receive packets with non-matching destination
      or source IPs. This is required if you plan to use this instance to forward
      routes.
    required: false
    type: bool
    aliases:
    - ip_forward
  deletion_protection:
    description:
    - Whether the resource should be protected against deletion.
    required: false
    type: bool
    version_added: 2.9
  disks:
    description:
    - An array of disks that are associated with the instances that are created from
      this template.
    required: false
    type: list
    suboptions:
      auto_delete:
        description:
        - Specifies whether the disk will be auto-deleted when the instance is deleted
          (but not when the disk is detached from the instance).
        - 'Tip: Disks should be set to autoDelete=true so that leftover disks are
          not left behind on machine deletion.'
        required: false
        type: bool
      boot:
        description:
        - Indicates that this is a boot disk. The virtual machine will use the first
          partition of the disk for its root filesystem.
        required: false
        type: bool
      device_name:
        description:
        - Specifies a unique device name of your choice that is reflected into the
          /dev/disk/by-id/google-* tree of a Linux operating system running within
          the instance. This name can be used to reference the device for mounting,
          resizing, and so on, from within the instance.
        required: false
        type: str
      disk_encryption_key:
        description:
        - Encrypts or decrypts a disk using a customer-supplied encryption key.
        required: false
        type: dict
        suboptions:
          raw_key:
            description:
            - Specifies a 256-bit customer-supplied encryption key, encoded in RFC
              4648 base64 to either encrypt or decrypt this resource.
            required: false
            type: str
          rsa_encrypted_key:
            description:
            - Specifies an RFC 4648 base64 encoded, RSA-wrapped 2048-bit customer-supplied
              encryption key to either encrypt or decrypt this resource.
            required: false
            type: str
      index:
        description:
        - Assigns a zero-based index to this disk, where 0 is reserved for the boot
          disk. For example, if you have many disks attached to an instance, each
          disk would have a unique index number. If not specified, the server will
          choose an appropriate value.
        required: false
        type: int
      initialize_params:
        description:
        - Specifies the parameters for a new disk that will be created alongside the
          new instance. Use initialization parameters to create boot disks or local
          SSDs attached to the new instance.
        required: false
        type: dict
        suboptions:
          disk_name:
            description:
            - Specifies the disk name. If not specified, the default is to use the
              name of the instance.
            required: false
            type: str
          disk_size_gb:
            description:
            - Specifies the size of the disk in base-2 GB.
            required: false
            type: int
          disk_type:
            description:
            - Reference to a disk type.
            - Specifies the disk type to use to create the instance.
            - If not specified, the default is pd-standard.
            required: false
            type: str
          source_image:
            description:
            - The source image to create this disk. When creating a new instance,
              one of initializeParams.sourceImage or disks.source is required. To
              create a disk with one of the public operating system images, specify
              the image by its family name.
            required: false
            type: str
            aliases:
            - image
            - image_family
          source_image_encryption_key:
            description:
            - The customer-supplied encryption key of the source image. Required if
              the source image is protected by a customer-supplied encryption key.
            - Instance templates do not store customer-supplied encryption keys, so
              you cannot create disks for instances in a managed instance group if
              the source images are encrypted with your own keys.
            required: false
            type: dict
            suboptions:
              raw_key:
                description:
                - Specifies a 256-bit customer-supplied encryption key, encoded in
                  RFC 4648 base64 to either encrypt or decrypt this resource.
                required: false
                type: str
      interface:
        description:
        - Specifies the disk interface to use for attaching this disk, which is either
          SCSI or NVME. The default is SCSI.
        - Persistent disks must always use SCSI and the request will fail if you attempt
          to attach a persistent disk in any other format than SCSI.
        - 'Some valid choices include: "SCSI", "NVME"'
        required: false
        type: str
      mode:
        description:
        - The mode in which to attach this disk, either READ_WRITE or READ_ONLY. If
          not specified, the default is to attach the disk in READ_WRITE mode.
        - 'Some valid choices include: "READ_WRITE", "READ_ONLY"'
        required: false
        type: str
      source:
        description:
        - Reference to a disk. When creating a new instance, one of initializeParams.sourceImage
          or disks.source is required.
        - If desired, you can also attach existing non-root persistent disks using
          this property. This field is only applicable for persistent disks.
        - 'This field represents a link to a Disk resource in GCP. It can be specified
          in two ways. First, you can place a dictionary with key ''selfLink'' and
          value of your resource''s selfLink Alternatively, you can add `register:
          name-of-resource` to a gcp_compute_disk task and then set this source field
          to "{{ name-of-resource }}"'
        required: false
        type: dict
      type:
        description:
        - Specifies the type of the disk, either SCRATCH or PERSISTENT. If not specified,
          the default is PERSISTENT.
        - 'Some valid choices include: "SCRATCH", "PERSISTENT"'
        required: false
        type: str
  guest_accelerators:
    description:
    - List of the type and count of accelerator cards attached to the instance .
    required: false
    type: list
    suboptions:
      accelerator_count:
        description:
        - The number of the guest accelerator cards exposed to this instance.
        required: false
        type: int
      accelerator_type:
        description:
        - Full or partial URL of the accelerator type resource to expose to this instance.
        required: false
        type: str
  hostname:
    description:
    - The hostname of the instance to be created. The specified hostname must be RFC1035
      compliant. If hostname is not specified, the default hostname is [INSTANCE_NAME].c.[PROJECT_ID].internal
      when using the global DNS, and [INSTANCE_NAME].[ZONE].c.[PROJECT_ID].internal
      when using zonal DNS.
    required: false
    type: str
    version_added: 2.9
  labels:
    description:
    - Labels to apply to this instance. A list of key->value pairs.
    required: false
    type: dict
    version_added: 2.9
  metadata:
    description:
    - The metadata key/value pairs to assign to instances that are created from this
      template. These pairs can consist of custom metadata or predefined keys.
    required: false
    type: dict
  machine_type:
    description:
    - A reference to a machine type which defines VM kind.
    required: false
    type: str
  min_cpu_platform:
    description:
    - Specifies a minimum CPU platform for the VM instance. Applicable values are
      the friendly names of CPU platforms .
    required: false
    type: str
  name:
    description:
    - The name of the resource, provided by the client when initially creating the
      resource. The resource name must be 1-63 characters long, and comply with RFC1035.
      Specifically, the name must be 1-63 characters long and match the regular expression
      `[a-z]([-a-z0-9]*[a-z0-9])?` which means the first character must be a lowercase
      letter, and all following characters must be a dash, lowercase letter, or digit,
      except the last character, which cannot be a dash.
    required: false
    type: str
  network_interfaces:
    description:
    - An array of configurations for this interface. This specifies how this interface
      is configured to interact with other network services, such as connecting to
      the internet. Only one network interface is supported per instance.
    required: false
    type: list
    suboptions:
      access_configs:
        description:
        - An array of configurations for this interface. Currently, only one access
          config, ONE_TO_ONE_NAT, is supported. If there are no accessConfigs specified,
          then this instance will have no external internet access.
        required: false
        type: list
        suboptions:
          name:
            description:
            - The name of this access configuration. The default and recommended name
              is External NAT but you can use any arbitrary string you would like.
              For example, My external IP or Network Access.
            required: true
            type: str
          nat_ip:
            description:
            - Reference to an address.
            - An external IP address associated with this instance.
            - Specify an unused static external IP address available to the project
              or leave this field undefined to use an IP from a shared ephemeral IP
              address pool. If you specify a static external IP address, it must live
              in the same region as the zone of the instance.
            - 'This field represents a link to a Address resource in GCP. It can be
              specified in two ways. First, you can place a dictionary with key ''address''
              and value of your resource''s address Alternatively, you can add `register:
              name-of-resource` to a gcp_compute_address task and then set this nat_ip
              field to "{{ name-of-resource }}"'
            required: false
            type: dict
          type:
            description:
            - The type of configuration. The default and only option is ONE_TO_ONE_NAT.
            - 'Some valid choices include: "ONE_TO_ONE_NAT"'
            required: true
            type: str
      alias_ip_ranges:
        description:
        - An array of alias IP ranges for this network interface. Can only be specified
          for network interfaces on subnet-mode networks.
        required: false
        type: list
        suboptions:
          ip_cidr_range:
            description:
            - The IP CIDR range represented by this alias IP range.
            - This IP CIDR range must belong to the specified subnetwork and cannot
              contain IP addresses reserved by system or used by other network interfaces.
              This range may be a single IP address (e.g. 10.2.3.4), a netmask (e.g.
              /24) or a CIDR format string (e.g. 10.1.2.0/24).
            required: false
            type: str
          subnetwork_range_name:
            description:
            - Optional subnetwork secondary range name specifying the secondary range
              from which to allocate the IP CIDR range for this alias IP range. If
              left unspecified, the primary range of the subnetwork will be used.
            required: false
            type: str
      network:
        description:
        - Specifies the title of an existing network. Not setting the network title
          will select the default network interface, which could have SSH already
          configured .
        - 'This field represents a link to a Network resource in GCP. It can be specified
          in two ways. First, you can place a dictionary with key ''selfLink'' and
          value of your resource''s selfLink Alternatively, you can add `register:
          name-of-resource` to a gcp_compute_network task and then set this network
          field to "{{ name-of-resource }}"'
        required: false
        type: dict
      network_ip:
        description:
        - An IPv4 internal network address to assign to the instance for this network
          interface. If not specified by the user, an unused internal IP is assigned
          by the system.
        required: false
        type: str
      subnetwork:
        description:
        - Reference to a VPC network.
        - If the network resource is in legacy mode, do not provide this property.
          If the network is in auto subnet mode, providing the subnetwork is optional.
          If the network is in custom subnet mode, then this field should be specified.
        - 'This field represents a link to a Subnetwork resource in GCP. It can be
          specified in two ways. First, you can place a dictionary with key ''selfLink''
          and value of your resource''s selfLink Alternatively, you can add `register:
          name-of-resource` to a gcp_compute_subnetwork task and then set this subnetwork
          field to "{{ name-of-resource }}"'
        required: false
        type: dict
  scheduling:
    description:
    - Sets the scheduling options for this instance.
    required: false
    type: dict
    suboptions:
      automatic_restart:
        description:
        - Specifies whether the instance should be automatically restarted if it is
          terminated by Compute Engine (not terminated by a user).
        - You can only set the automatic restart option for standard instances. Preemptible
          instances cannot be automatically restarted.
        required: false
        type: bool
      on_host_maintenance:
        description:
        - Defines the maintenance behavior for this instance. For standard instances,
          the default behavior is MIGRATE. For preemptible instances, the default
          and only possible behavior is TERMINATE.
        - For more information, see Setting Instance Scheduling Options.
        required: false
        type: str
      preemptible:
        description:
        - Defines whether the instance is preemptible. This can only be set during
          instance creation, it cannot be set or changed after the instance has been
          created.
        required: false
        type: bool
  service_accounts:
    description:
    - A list of service accounts, with their specified scopes, authorized for this
      instance. Only one service account per VM instance is supported.
    required: false
    type: list
    suboptions:
      email:
        description:
        - Email address of the service account.
        required: false
        type: str
      scopes:
        description:
        - The list of scopes to be made available for this service account.
        required: false
        type: list
  shielded_instance_config:
    description:
    - Configuration for various parameters related to shielded instances.
    required: false
    type: dict
    version_added: 2.9
    suboptions:
      enable_secure_boot:
        description:
        - Defines whether the instance has Secure Boot enabled.
        required: false
        type: bool
      enable_vtpm:
        description:
        - Defines whether the instance has the vTPM enabled.
        required: false
        type: bool
      enable_integrity_monitoring:
        description:
        - Defines whether the instance has integrity monitoring enabled.
        required: false
        type: bool
  status:
    description:
    - 'The status of the instance. One of the following values: PROVISIONING, STAGING,
      RUNNING, STOPPING, SUSPENDING, SUSPENDED, and TERMINATED.'
    - As a user, use RUNNING to keep a machine "on" and TERMINATED to turn a machine
      off .
    - 'Some valid choices include: "PROVISIONING", "STAGING", "RUNNING", "STOPPING",
      "SUSPENDING", "SUSPENDED", "TERMINATED"'
    required: false
    type: str
    version_added: 2.8
  tags:
    description:
    - A list of tags to apply to this instance. Tags are used to identify valid sources
      or targets for network firewalls and are specified by the client during instance
      creation. The tags can be later modified by the setTags method. Each tag within
      the list must comply with RFC1035.
    required: false
    type: dict
    suboptions:
      fingerprint:
        description:
        - Specifies a fingerprint for this request, which is essentially a hash of
          the metadata's contents and used for optimistic locking.
        - The fingerprint is initially generated by Compute Engine and changes after
          every request to modify or update metadata. You must always provide an up-to-date
          fingerprint hash in order to update or change metadata.
        required: false
        type: str
      items:
        description:
        - An array of tags. Each tag must be 1-63 characters long, and comply with
          RFC1035.
        required: false
        type: list
  zone:
    description:
    - A reference to the zone where the machine resides.
    required: true
    type: str
extends_documentation_fragment: gcp
'''

EXAMPLES = '''
- name: create a disk
  gcp_compute_disk:
    name: disk-instance
    size_gb: 50
    source_image: projects/ubuntu-os-cloud/global/images/family/ubuntu-1604-lts
    zone: us-central1-a
    project: "{{ gcp_project }}"
    auth_kind: "{{ gcp_cred_kind }}"
    service_account_file: "{{ gcp_cred_file }}"
    state: present
  register: disk

- name: create a network
  gcp_compute_network:
    name: network-instance
    project: "{{ gcp_project }}"
    auth_kind: "{{ gcp_cred_kind }}"
    service_account_file: "{{ gcp_cred_file }}"
    state: present
  register: network

- name: create a address
  gcp_compute_address:
    name: address-instance
    region: us-central1
    project: "{{ gcp_project }}"
    auth_kind: "{{ gcp_cred_kind }}"
    service_account_file: "{{ gcp_cred_file }}"
    state: present
  register: address

- name: create a instance
  gcp_compute_instance:
    name: test_object
    machine_type: n1-standard-1
    disks:
    - auto_delete: 'true'
      boot: 'true'
      source: "{{ disk }}"
    metadata:
      startup-script-url: gs:://graphite-playground/bootstrap.sh
      cost-center: '12345'
    labels:
      environment: production
    network_interfaces:
    - network: "{{ network }}"
      access_configs:
      - name: External NAT
        nat_ip: "{{ address }}"
        type: ONE_TO_ONE_NAT
    zone: us-central1-a
    project: test_project
    auth_kind: serviceaccount
    service_account_file: "/tmp/auth.pem"
    state: present
'''

RETURN = '''
canIpForward:
  description:
  - Allows this instance to send and receive packets with non-matching destination
    or source IPs. This is required if you plan to use this instance to forward routes.
  returned: success
  type: bool
cpuPlatform:
  description:
  - The CPU platform used by this instance.
  returned: success
  type: str
creationTimestamp:
  description:
  - Creation timestamp in RFC3339 text format.
  returned: success
  type: str
deletionProtection:
  description:
  - Whether the resource should be protected against deletion.
  returned: success
  type: bool
disks:
  description:
  - An array of disks that are associated with the instances that are created from
    this template.
  returned: success
  type: complex
  contains:
    autoDelete:
      description:
      - Specifies whether the disk will be auto-deleted when the instance is deleted
        (but not when the disk is detached from the instance).
      - 'Tip: Disks should be set to autoDelete=true so that leftover disks are not
        left behind on machine deletion.'
      returned: success
      type: bool
    boot:
      description:
      - Indicates that this is a boot disk. The virtual machine will use the first
        partition of the disk for its root filesystem.
      returned: success
      type: bool
    deviceName:
      description:
      - Specifies a unique device name of your choice that is reflected into the /dev/disk/by-id/google-*
        tree of a Linux operating system running within the instance. This name can
        be used to reference the device for mounting, resizing, and so on, from within
        the instance.
      returned: success
      type: str
    diskEncryptionKey:
      description:
      - Encrypts or decrypts a disk using a customer-supplied encryption key.
      returned: success
      type: complex
      contains:
        rawKey:
          description:
          - Specifies a 256-bit customer-supplied encryption key, encoded in RFC 4648
            base64 to either encrypt or decrypt this resource.
          returned: success
          type: str
        rsaEncryptedKey:
          description:
          - Specifies an RFC 4648 base64 encoded, RSA-wrapped 2048-bit customer-supplied
            encryption key to either encrypt or decrypt this resource.
          returned: success
          type: str
        sha256:
          description:
          - The RFC 4648 base64 encoded SHA-256 hash of the customer-supplied encryption
            key that protects this resource.
          returned: success
          type: str
    index:
      description:
      - Assigns a zero-based index to this disk, where 0 is reserved for the boot
        disk. For example, if you have many disks attached to an instance, each disk
        would have a unique index number. If not specified, the server will choose
        an appropriate value.
      returned: success
      type: int
    initializeParams:
      description:
      - Specifies the parameters for a new disk that will be created alongside the
        new instance. Use initialization parameters to create boot disks or local
        SSDs attached to the new instance.
      returned: success
      type: complex
      contains:
        diskName:
          description:
          - Specifies the disk name. If not specified, the default is to use the name
            of the instance.
          returned: success
          type: str
        diskSizeGb:
          description:
          - Specifies the size of the disk in base-2 GB.
          returned: success
          type: int
        diskType:
          description:
          - Reference to a disk type.
          - Specifies the disk type to use to create the instance.
          - If not specified, the default is pd-standard.
          returned: success
          type: str
        sourceImage:
          description:
          - The source image to create this disk. When creating a new instance, one
            of initializeParams.sourceImage or disks.source is required. To create
            a disk with one of the public operating system images, specify the image
            by its family name.
          returned: success
          type: str
        sourceImageEncryptionKey:
          description:
          - The customer-supplied encryption key of the source image. Required if
            the source image is protected by a customer-supplied encryption key.
          - Instance templates do not store customer-supplied encryption keys, so
            you cannot create disks for instances in a managed instance group if the
            source images are encrypted with your own keys.
          returned: success
          type: complex
          contains:
            rawKey:
              description:
              - Specifies a 256-bit customer-supplied encryption key, encoded in RFC
                4648 base64 to either encrypt or decrypt this resource.
              returned: success
              type: str
            sha256:
              description:
              - The RFC 4648 base64 encoded SHA-256 hash of the customer-supplied
                encryption key that protects this resource.
              returned: success
              type: str
    interface:
      description:
      - Specifies the disk interface to use for attaching this disk, which is either
        SCSI or NVME. The default is SCSI.
      - Persistent disks must always use SCSI and the request will fail if you attempt
        to attach a persistent disk in any other format than SCSI.
      returned: success
      type: str
    mode:
      description:
      - The mode in which to attach this disk, either READ_WRITE or READ_ONLY. If
        not specified, the default is to attach the disk in READ_WRITE mode.
      returned: success
      type: str
    source:
      description:
      - Reference to a disk. When creating a new instance, one of initializeParams.sourceImage
        or disks.source is required.
      - If desired, you can also attach existing non-root persistent disks using this
        property. This field is only applicable for persistent disks.
      returned: success
      type: dict
    type:
      description:
      - Specifies the type of the disk, either SCRATCH or PERSISTENT. If not specified,
        the default is PERSISTENT.
      returned: success
      type: str
guestAccelerators:
  description:
  - List of the type and count of accelerator cards attached to the instance .
  returned: success
  type: complex
  contains:
    acceleratorCount:
      description:
      - The number of the guest accelerator cards exposed to this instance.
      returned: success
      type: int
    acceleratorType:
      description:
      - Full or partial URL of the accelerator type resource to expose to this instance.
      returned: success
      type: str
hostname:
  description:
  - The hostname of the instance to be created. The specified hostname must be RFC1035
    compliant. If hostname is not specified, the default hostname is [INSTANCE_NAME].c.[PROJECT_ID].internal
    when using the global DNS, and [INSTANCE_NAME].[ZONE].c.[PROJECT_ID].internal
    when using zonal DNS.
  returned: success
  type: str
id:
  description:
  - The unique identifier for the resource. This identifier is defined by the server.
  returned: success
  type: int
labelFingerprint:
  description:
  - The fingerprint used for optimistic locking of this resource. Used internally
    during updates.
  returned: success
  type: str
labels:
  description:
  - Labels to apply to this instance. A list of key->value pairs.
  returned: success
  type: dict
metadata:
  description:
  - The metadata key/value pairs to assign to instances that are created from this
    template. These pairs can consist of custom metadata or predefined keys.
  returned: success
  type: dict
machineType:
  description:
  - A reference to a machine type which defines VM kind.
  returned: success
  type: str
minCpuPlatform:
  description:
  - Specifies a minimum CPU platform for the VM instance. Applicable values are the
    friendly names of CPU platforms .
  returned: success
  type: str
name:
  description:
  - The name of the resource, provided by the client when initially creating the resource.
    The resource name must be 1-63 characters long, and comply with RFC1035. Specifically,
    the name must be 1-63 characters long and match the regular expression `[a-z]([-a-z0-9]*[a-z0-9])?`
    which means the first character must be a lowercase letter, and all following
    characters must be a dash, lowercase letter, or digit, except the last character,
    which cannot be a dash.
  returned: success
  type: str
networkInterfaces:
  description:
  - An array of configurations for this interface. This specifies how this interface
    is configured to interact with other network services, such as connecting to the
    internet. Only one network interface is supported per instance.
  returned: success
  type: complex
  contains:
    accessConfigs:
      description:
      - An array of configurations for this interface. Currently, only one access
        config, ONE_TO_ONE_NAT, is supported. If there are no accessConfigs specified,
        then this instance will have no external internet access.
      returned: success
      type: complex
      contains:
        name:
          description:
          - The name of this access configuration. The default and recommended name
            is External NAT but you can use any arbitrary string you would like. For
            example, My external IP or Network Access.
          returned: success
          type: str
        natIP:
          description:
          - Reference to an address.
          - An external IP address associated with this instance.
          - Specify an unused static external IP address available to the project
            or leave this field undefined to use an IP from a shared ephemeral IP
            address pool. If you specify a static external IP address, it must live
            in the same region as the zone of the instance.
          returned: success
          type: dict
        type:
          description:
          - The type of configuration. The default and only option is ONE_TO_ONE_NAT.
          returned: success
          type: str
    aliasIpRanges:
      description:
      - An array of alias IP ranges for this network interface. Can only be specified
        for network interfaces on subnet-mode networks.
      returned: success
      type: complex
      contains:
        ipCidrRange:
          description:
          - The IP CIDR range represented by this alias IP range.
          - This IP CIDR range must belong to the specified subnetwork and cannot
            contain IP addresses reserved by system or used by other network interfaces.
            This range may be a single IP address (e.g. 10.2.3.4), a netmask (e.g.
            /24) or a CIDR format string (e.g. 10.1.2.0/24).
          returned: success
          type: str
        subnetworkRangeName:
          description:
          - Optional subnetwork secondary range name specifying the secondary range
            from which to allocate the IP CIDR range for this alias IP range. If left
            unspecified, the primary range of the subnetwork will be used.
          returned: success
          type: str
    name:
      description:
      - The name of the network interface, generated by the server. For network devices,
        these are eth0, eth1, etc .
      returned: success
      type: str
    network:
      description:
      - Specifies the title of an existing network. Not setting the network title
        will select the default network interface, which could have SSH already configured
        .
      returned: success
      type: dict
    networkIP:
      description:
      - An IPv4 internal network address to assign to the instance for this network
        interface. If not specified by the user, an unused internal IP is assigned
        by the system.
      returned: success
      type: str
    subnetwork:
      description:
      - Reference to a VPC network.
      - If the network resource is in legacy mode, do not provide this property. If
        the network is in auto subnet mode, providing the subnetwork is optional.
        If the network is in custom subnet mode, then this field should be specified.
      returned: success
      type: dict
scheduling:
  description:
  - Sets the scheduling options for this instance.
  returned: success
  type: complex
  contains:
    automaticRestart:
      description:
      - Specifies whether the instance should be automatically restarted if it is
        terminated by Compute Engine (not terminated by a user).
      - You can only set the automatic restart option for standard instances. Preemptible
        instances cannot be automatically restarted.
      returned: success
      type: bool
    onHostMaintenance:
      description:
      - Defines the maintenance behavior for this instance. For standard instances,
        the default behavior is MIGRATE. For preemptible instances, the default and
        only possible behavior is TERMINATE.
      - For more information, see Setting Instance Scheduling Options.
      returned: success
      type: str
    preemptible:
      description:
      - Defines whether the instance is preemptible. This can only be set during instance
        creation, it cannot be set or changed after the instance has been created.
      returned: success
      type: bool
serviceAccounts:
  description:
  - A list of service accounts, with their specified scopes, authorized for this instance.
    Only one service account per VM instance is supported.
  returned: success
  type: complex
  contains:
    email:
      description:
      - Email address of the service account.
      returned: success
      type: str
    scopes:
      description:
      - The list of scopes to be made available for this service account.
      returned: success
      type: list
shieldedInstanceConfig:
  description:
  - Configuration for various parameters related to shielded instances.
  returned: success
  type: complex
  contains:
    enableSecureBoot:
      description:
      - Defines whether the instance has Secure Boot enabled.
      returned: success
      type: bool
    enableVtpm:
      description:
      - Defines whether the instance has the vTPM enabled.
      returned: success
      type: bool
    enableIntegrityMonitoring:
      description:
      - Defines whether the instance has integrity monitoring enabled.
      returned: success
      type: bool
status:
  description:
  - 'The status of the instance. One of the following values: PROVISIONING, STAGING,
    RUNNING, STOPPING, SUSPENDING, SUSPENDED, and TERMINATED.'
  - As a user, use RUNNING to keep a machine "on" and TERMINATED to turn a machine
    off .
  returned: success
  type: str
statusMessage:
  description:
  - An optional, human-readable explanation of the status.
  returned: success
  type: str
tags:
  description:
  - A list of tags to apply to this instance. Tags are used to identify valid sources
    or targets for network firewalls and are specified by the client during instance
    creation. The tags can be later modified by the setTags method. Each tag within
    the list must comply with RFC1035.
  returned: success
  type: complex
  contains:
    fingerprint:
      description:
      - Specifies a fingerprint for this request, which is essentially a hash of the
        metadata's contents and used for optimistic locking.
      - The fingerprint is initially generated by Compute Engine and changes after
        every request to modify or update metadata. You must always provide an up-to-date
        fingerprint hash in order to update or change metadata.
      returned: success
      type: str
    items:
      description:
      - An array of tags. Each tag must be 1-63 characters long, and comply with RFC1035.
      returned: success
      type: list
zone:
  description:
  - A reference to the zone where the machine resides.
  returned: success
  type: str
'''

################################################################################
# Imports
################################################################################

from ansible.module_utils.gcp_utils import navigate_hash, GcpSession, GcpModule, GcpRequest, remove_nones_from_dict, replace_resource_dict
import json
import re
import time

################################################################################
# Main
################################################################################


def main():
    """Main function"""

    module = GcpModule(
        argument_spec=dict(
            state=dict(default='present', choices=['present', 'absent'], type='str'),
            can_ip_forward=dict(type='bool', aliases=['ip_forward']),
            deletion_protection=dict(type='bool'),
            disks=dict(
                type='list',
                elements='dict',
                options=dict(
                    auto_delete=dict(type='bool'),
                    boot=dict(type='bool'),
                    device_name=dict(type='str'),
                    disk_encryption_key=dict(type='dict', options=dict(raw_key=dict(type='str'), rsa_encrypted_key=dict(type='str'))),
                    index=dict(type='int'),
                    initialize_params=dict(
                        type='dict',
                        options=dict(
                            disk_name=dict(type='str'),
                            disk_size_gb=dict(type='int'),
                            disk_type=dict(type='str'),
                            source_image=dict(type='str', aliases=['image', 'image_family']),
                            source_image_encryption_key=dict(type='dict', options=dict(raw_key=dict(type='str'))),
                        ),
                    ),
                    interface=dict(type='str'),
                    mode=dict(type='str'),
                    source=dict(type='dict'),
                    type=dict(type='str'),
                ),
            ),
            guest_accelerators=dict(type='list', elements='dict', options=dict(accelerator_count=dict(type='int'), accelerator_type=dict(type='str'))),
            hostname=dict(type='str'),
            labels=dict(type='dict'),
            metadata=dict(type='dict'),
            machine_type=dict(type='str'),
            min_cpu_platform=dict(type='str'),
            name=dict(type='str'),
            network_interfaces=dict(
                type='list',
                elements='dict',
                options=dict(
                    access_configs=dict(
                        type='list',
                        elements='dict',
                        options=dict(name=dict(required=True, type='str'), nat_ip=dict(type='dict'), type=dict(required=True, type='str')),
                    ),
                    alias_ip_ranges=dict(type='list', elements='dict', options=dict(ip_cidr_range=dict(type='str'), subnetwork_range_name=dict(type='str'))),
                    network=dict(type='dict'),
                    network_ip=dict(type='str'),
                    subnetwork=dict(type='dict'),
                ),
            ),
            scheduling=dict(
                type='dict', options=dict(automatic_restart=dict(type='bool'), on_host_maintenance=dict(type='str'), preemptible=dict(type='bool'))
            ),
            service_accounts=dict(type='list', elements='dict', options=dict(email=dict(type='str'), scopes=dict(type='list', elements='str'))),
            shielded_instance_config=dict(
                type='dict', options=dict(enable_secure_boot=dict(type='bool'), enable_vtpm=dict(type='bool'), enable_integrity_monitoring=dict(type='bool'))
            ),
            status=dict(type='str'),
            tags=dict(type='dict', options=dict(fingerprint=dict(type='str'), items=dict(type='list', elements='str'))),
            zone=dict(required=True, type='str'),
        )
    )

    if not module.params['scopes']:
        module.params['scopes'] = ['https://www.googleapis.com/auth/compute']

    state = module.params['state']
    kind = 'compute#instance'

    fetch = fetch_resource(module, self_link(module), kind)
    changed = False

    if fetch:
        if state == 'present':
            if is_different(module, fetch):
                update(module, self_link(module), kind, fetch)
                fetch = fetch_resource(module, self_link(module), kind)
                changed = True
        else:
            delete(module, self_link(module), kind)
            fetch = {}
            changed = True
    else:
        if state == 'present':
            fetch = create(module, collection(module), kind)
            changed = True
        else:
            fetch = {}

    if fetch:
        instance = InstancePower(module, fetch.get('status'))
        instance.run()
        if module.params.get('status'):
            fetch.update({'status': module.params['status']})
    fetch.update({'changed': changed})

    module.exit_json(**fetch)


def create(module, link, kind):
    auth = GcpSession(module, 'compute')
    return wait_for_operation(module, auth.post(link, resource_to_request(module)))


def update(module, link, kind, fetch):
    update_fields(module, resource_to_request(module), response_to_hash(module, fetch))
    return fetch_resource(module, self_link(module), kind)


def update_fields(module, request, response):
    if response.get('deletionProtection') != request.get('deletionProtection'):
        deletion_protection_update(module, request, response)
    if response.get('labels') != request.get('labels'):
        label_fingerprint_update(module, request, response)
    if response.get('machineType') != request.get('machineType'):
        machine_type_update(module, request, response)
    if response.get('shieldedInstanceConfig') != request.get('shieldedInstanceConfig'):
        shielded_instance_config_update(module, request, response)


def label_fingerprint_update(module, request, response):
    auth = GcpSession(module, 'compute')
    auth.post(
        ''.join(["https://www.googleapis.com/compute/v1/", "projects/{project}/zones/{zone}/instances/{name}/setLabels"]).format(**module.params),
        {u'labelFingerprint': response.get('labelFingerprint'), u'labels': module.params.get('labels')},
    )


def machine_type_update(module, request, response):
    auth = GcpSession(module, 'compute')
    auth.post(
        ''.join(["https://www.googleapis.com/compute/v1/", "projects/{project}/zones/{zone}/instances/{name}/setMachineType"]).format(**module.params),
        {u'machineType': machine_type_selflink(module.params.get('machine_type'), module.params)},
    )


def delete(module, link, kind):
    auth = GcpSession(module, 'compute')
    return wait_for_operation(module, auth.delete(link))


def resource_to_request(module):
    request = {
        u'kind': 'compute#instance',
        u'canIpForward': module.params.get('can_ip_forward'),
        u'deletionProtection': module.params.get('deletion_protection'),
        u'disks': InstanceDisksArray(module.params.get('disks', []), module).to_request(),
        u'guestAccelerators': InstanceGuestacceleratorsArray(module.params.get('guest_accelerators', []), module).to_request(),
        u'hostname': module.params.get('hostname'),
        u'labels': module.params.get('labels'),
        u'metadata': module.params.get('metadata'),
        u'machineType': machine_type_selflink(module.params.get('machine_type'), module.params),
        u'minCpuPlatform': module.params.get('min_cpu_platform'),
        u'name': module.params.get('name'),
        u'networkInterfaces': InstanceNetworkinterfacesArray(module.params.get('network_interfaces', []), module).to_request(),
        u'scheduling': InstanceScheduling(module.params.get('scheduling', {}), module).to_request(),
        u'serviceAccounts': InstanceServiceaccountsArray(module.params.get('service_accounts', []), module).to_request(),
        u'shieldedInstanceConfig': InstanceShieldedinstanceconfig(module.params.get('shielded_instance_config', {}), module).to_request(),
        u'status': module.params.get('status'),
        u'tags': InstanceTags(module.params.get('tags', {}), module).to_request(),
    }
    request = encode_request(request, module)
    return_vals = {}
    for k, v in request.items():
        if v or v is False:
            return_vals[k] = v

    return return_vals


def fetch_resource(module, link, kind, allow_not_found=True):
    auth = GcpSession(module, 'compute')
    return return_if_object(module, auth.get(link), kind, allow_not_found)


def self_link(module):
    return "https://www.googleapis.com/compute/v1/projects/{project}/zones/{zone}/instances/{name}".format(**module.params)


def collection(module):
    return "https://www.googleapis.com/compute/v1/projects/{project}/zones/{zone}/instances".format(**module.params)


def return_if_object(module, response, kind, allow_not_found=False):
    # If not found, return nothing.
    if allow_not_found and response.status_code == 404:
        return None

    # If no content, return nothing.
    if response.status_code == 204:
        return None

    try:
        module.raise_for_status(response)
        result = response.json()
    except getattr(json.decoder, 'JSONDecodeError', ValueError):
        module.fail_json(msg="Invalid JSON response with error: %s" % response.text)

    result = decode_response(result, module)

    if navigate_hash(result, ['error', 'errors']):
        module.fail_json(msg=navigate_hash(result, ['error', 'errors']))

    return result


def is_different(module, response):
    request = resource_to_request(module)
    response = response_to_hash(module, response)
    request = decode_response(request, module)

    # Remove all output-only from response.
    response_vals = {}
    for k, v in response.items():
        if k in request:
            response_vals[k] = v

    request_vals = {}
    for k, v in request.items():
        if k in response:
            request_vals[k] = v

    return GcpRequest(request_vals) != GcpRequest(response_vals)


# Remove unnecessary properties from the response.
# This is for doing comparisons with Ansible's current parameters.
def response_to_hash(module, response):
    return {
        u'canIpForward': response.get(u'canIpForward'),
        u'cpuPlatform': response.get(u'cpuPlatform'),
        u'creationTimestamp': response.get(u'creationTimestamp'),
        u'deletionProtection': response.get(u'deletionProtection'),
        u'disks': InstanceDisksArray(module.params.get('disks', []), module).to_request(),
        u'guestAccelerators': InstanceGuestacceleratorsArray(response.get(u'guestAccelerators', []), module).from_response(),
        u'hostname': response.get(u'hostname'),
        u'id': response.get(u'id'),
        u'labelFingerprint': response.get(u'labelFingerprint'),
        u'labels': response.get(u'labels'),
        u'metadata': response.get(u'metadata'),
        u'machineType': response.get(u'machineType'),
        u'minCpuPlatform': response.get(u'minCpuPlatform'),
        u'name': response.get(u'name'),
        u'networkInterfaces': InstanceNetworkinterfacesArray(response.get(u'networkInterfaces', []), module).from_response(),
        u'scheduling': InstanceScheduling(response.get(u'scheduling', {}), module).from_response(),
        u'serviceAccounts': InstanceServiceaccountsArray(response.get(u'serviceAccounts', []), module).from_response(),
        u'shieldedInstanceConfig': InstanceShieldedinstanceconfig(response.get(u'shieldedInstanceConfig', {}), module).from_response(),
        u'status': response.get(u'status'),
        u'statusMessage': response.get(u'statusMessage'),
        u'tags': InstanceTags(response.get(u'tags', {}), module).from_response(),
    }


def disk_type_selflink(name, params):
    if name is None:
        return
    url = r"https://www.googleapis.com/compute/v1/projects/.*/zones/.*/diskTypes/.*"
    if not re.match(url, name):
        name = "https://www.googleapis.com/compute/v1/projects/{project}/zones/{zone}/diskTypes/%s".format(**params) % name
    return name


def machine_type_selflink(name, params):
    if name is None:
        return
    url = r"https://www.googleapis.com/compute/v1/projects/.*/zones/.*/machineTypes/.*"
    if not re.match(url, name):
        name = "https://www.googleapis.com/compute/v1/projects/{project}/zones/{zone}/machineTypes/%s".format(**params) % name
    return name


def async_op_url(module, extra_data=None):
    if extra_data is None:
        extra_data = {}
    url = "https://www.googleapis.com/compute/v1/projects/{project}/zones/{zone}/operations/{op_id}"
    combined = extra_data.copy()
    combined.update(module.params)
    return url.format(**combined)


def wait_for_operation(module, response):
    op_result = return_if_object(module, response, 'compute#operation')
    if op_result is None:
        return {}
    status = navigate_hash(op_result, ['status'])
    wait_done = wait_for_completion(status, op_result, module)
    response = fetch_resource(module, navigate_hash(wait_done, ['targetLink']), 'compute#instance')
    if response:
        return decode_response(response, module)
    else:
        return {}


def wait_for_completion(status, op_result, module):
    op_id = navigate_hash(op_result, ['name'])
    op_uri = async_op_url(module, {'op_id': op_id})
    while status != 'DONE':
        raise_if_errors(op_result, ['error', 'errors'], module)
        time.sleep(1.0)
        op_result = fetch_resource(module, op_uri, 'compute#operation', False)
        status = navigate_hash(op_result, ['status'])
    return op_result


def raise_if_errors(response, err_path, module):
    errors = navigate_hash(response, err_path)
    if errors is not None:
        module.fail_json(msg=errors)


def encode_request(request, module):
    if 'metadata' in request and request['metadata'] is not None:
        request['metadata'] = metadata_encoder(request['metadata'])
    return request


def decode_response(response, module):
    if 'metadata' in response and response['metadata'] is not None:
        response['metadata'] = metadata_decoder(response['metadata'])
    return response


# TODO(alexstephen): Implement updating metadata on existing resources.

# Expose instance 'metadata' as a simple name/value pair hash. However the API
# defines metadata as a NestedObject with the following layout:
#
# metadata {
#   fingerprint: 'hash-of-last-metadata'
#   items: [
#     {
#       key: 'metadata1-key'
#       value: 'metadata1-value'
#     },
#     ...
#   ]
# }
#
def metadata_encoder(metadata):
    metadata_new = []
    for key in metadata:
        value = metadata[key]
        metadata_new.append({"key": key, "value": value})
    return {'items': metadata_new}


# Map metadata.items[]{key:,value:} => metadata[key]=value
def metadata_decoder(metadata):
    items = {}
    if 'items' in metadata:
        metadata_items = metadata['items']
        for item in metadata_items:
            items[item['key']] = item['value']
    return items


class InstancePower(object):
    def __init__(self, module, current_status):
        self.module = module
        self.current_status = current_status
        self.desired_status = self.module.params.get('status')

    def run(self):
        # GcpRequest handles unicode text handling
        if GcpRequest({'status': self.current_status}) == GcpRequest({'status': self.desired_status}):
            return
        elif self.desired_status == 'RUNNING':
            self.start()
        elif self.desired_status == 'TERMINATED':
            self.stop()
        elif self.desired_status == 'SUSPENDED':
            self.module.fail_json(msg="Instances cannot be suspended using Ansible")

    def start(self):
        auth = GcpSession(self.module, 'compute')
        wait_for_operation(self.module, auth.post(self._start_url()))

    def stop(self):
        auth = GcpSession(self.module, 'compute')
        wait_for_operation(self.module, auth.post(self._stop_url()))

    def _start_url(self):
        return "https://www.googleapis.com/compute/v1/projects/{project}/zones/{zone}/instances/{name}/start".format(**self.module.params)

    def _stop_url(self):
        return "https://www.googleapis.com/compute/v1/projects/{project}/zones/{zone}/instances/{name}/stop".format(**self.module.params)


def deletion_protection_update(module, request, response):
    auth = GcpSession(module, 'compute')
    auth.post(
        ''.join(
            [
                "https://www.googleapis.com/compute/v1/",
                "projects/{project}/zones/{zone}/instances/{name}/setDeletionProtection?deletionProtection={deletionProtection}",
            ]
        ).format(**module.params),
        {},
    )


def shielded_instance_config_update(module, request, response):
    auth = GcpSession(module, 'compute')
    auth.post(
        ''.join(["https://www.googleapis.com/compute/v1/", "projects/{project}/zones/{zone}/instances/{name}/updateShieldedInstanceConfig"]).format(
            **module.params
        ),
        {
            u'enableSecureBoot': navigate_hash(module.params, ['shielded_instance_config', 'enable_secure_boot']),
            u'enableVtpm': navigate_hash(module.params, ['shielded_instance_config', 'enable_vtpm']),
            u'enableIntegrityMonitoring': navigate_hash(module.params, ['shielded_instance_config', 'enable_integrity_monitoring']),
        },
    )


class InstanceDisksArray(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = []

    def to_request(self):
        items = []
        for item in self.request:
            items.append(self._request_for_item(item))
        return items

    def from_response(self):
        items = []
        for item in self.request:
            items.append(self._response_from_item(item))
        return items

    def _request_for_item(self, item):
        return remove_nones_from_dict(
            {
                u'autoDelete': item.get('auto_delete'),
                u'boot': item.get('boot'),
                u'deviceName': item.get('device_name'),
                u'diskEncryptionKey': InstanceDiskencryptionkey(item.get('disk_encryption_key', {}), self.module).to_request(),
                u'index': item.get('index'),
                u'initializeParams': InstanceInitializeparams(item.get('initialize_params', {}), self.module).to_request(),
                u'interface': item.get('interface'),
                u'mode': item.get('mode'),
                u'source': replace_resource_dict(item.get(u'source', {}), 'selfLink'),
                u'type': item.get('type'),
            }
        )

    def _response_from_item(self, item):
        return remove_nones_from_dict(
            {
                u'autoDelete': item.get(u'autoDelete'),
                u'boot': item.get(u'boot'),
                u'deviceName': item.get(u'deviceName'),
                u'diskEncryptionKey': InstanceDiskencryptionkey(item.get(u'diskEncryptionKey', {}), self.module).from_response(),
                u'index': item.get(u'index'),
                u'initializeParams': InstanceInitializeparams(self.module.params.get('initialize_params', {}), self.module).to_request(),
                u'interface': item.get(u'interface'),
                u'mode': item.get(u'mode'),
                u'source': item.get(u'source'),
                u'type': item.get(u'type'),
            }
        )


class InstanceDiskencryptionkey(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = {}

    def to_request(self):
        return remove_nones_from_dict({u'rawKey': self.request.get('raw_key'), u'rsaEncryptedKey': self.request.get('rsa_encrypted_key')})

    def from_response(self):
        return remove_nones_from_dict({u'rawKey': self.request.get(u'rawKey'), u'rsaEncryptedKey': self.request.get(u'rsaEncryptedKey')})


class InstanceInitializeparams(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = {}

    def to_request(self):
        return remove_nones_from_dict(
            {
                u'diskName': self.request.get('disk_name'),
                u'diskSizeGb': self.request.get('disk_size_gb'),
                u'diskType': disk_type_selflink(self.request.get('disk_type'), self.module.params),
                u'sourceImage': self.request.get('source_image'),
                u'sourceImageEncryptionKey': InstanceSourceimageencryptionkey(self.request.get('source_image_encryption_key', {}), self.module).to_request(),
            }
        )

    def from_response(self):
        return remove_nones_from_dict(
            {
                u'diskName': self.request.get(u'diskName'),
                u'diskSizeGb': self.request.get(u'diskSizeGb'),
                u'diskType': self.request.get(u'diskType'),
                u'sourceImage': self.request.get(u'sourceImage'),
                u'sourceImageEncryptionKey': InstanceSourceimageencryptionkey(self.request.get(u'sourceImageEncryptionKey', {}), self.module).from_response(),
            }
        )


class InstanceSourceimageencryptionkey(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = {}

    def to_request(self):
        return remove_nones_from_dict({u'rawKey': self.request.get('raw_key')})

    def from_response(self):
        return remove_nones_from_dict({u'rawKey': self.request.get(u'rawKey')})


class InstanceGuestacceleratorsArray(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = []

    def to_request(self):
        items = []
        for item in self.request:
            items.append(self._request_for_item(item))
        return items

    def from_response(self):
        items = []
        for item in self.request:
            items.append(self._response_from_item(item))
        return items

    def _request_for_item(self, item):
        return remove_nones_from_dict({u'acceleratorCount': item.get('accelerator_count'), u'acceleratorType': item.get('accelerator_type')})

    def _response_from_item(self, item):
        return remove_nones_from_dict({u'acceleratorCount': item.get(u'acceleratorCount'), u'acceleratorType': item.get(u'acceleratorType')})


class InstanceNetworkinterfacesArray(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = []

    def to_request(self):
        items = []
        for item in self.request:
            items.append(self._request_for_item(item))
        return items

    def from_response(self):
        items = []
        for item in self.request:
            items.append(self._response_from_item(item))
        return items

    def _request_for_item(self, item):
        return remove_nones_from_dict(
            {
                u'accessConfigs': InstanceAccessconfigsArray(item.get('access_configs', []), self.module).to_request(),
                u'aliasIpRanges': InstanceAliasiprangesArray(item.get('alias_ip_ranges', []), self.module).to_request(),
                u'network': replace_resource_dict(item.get(u'network', {}), 'selfLink'),
                u'networkIP': item.get('network_ip'),
                u'subnetwork': replace_resource_dict(item.get(u'subnetwork', {}), 'selfLink'),
            }
        )

    def _response_from_item(self, item):
        return remove_nones_from_dict(
            {
                u'accessConfigs': InstanceAccessconfigsArray(item.get(u'accessConfigs', []), self.module).from_response(),
                u'aliasIpRanges': InstanceAliasiprangesArray(item.get(u'aliasIpRanges', []), self.module).from_response(),
                u'network': item.get(u'network'),
                u'networkIP': item.get(u'networkIP'),
                u'subnetwork': item.get(u'subnetwork'),
            }
        )


class InstanceAccessconfigsArray(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = []

    def to_request(self):
        items = []
        for item in self.request:
            items.append(self._request_for_item(item))
        return items

    def from_response(self):
        items = []
        for item in self.request:
            items.append(self._response_from_item(item))
        return items

    def _request_for_item(self, item):
        return remove_nones_from_dict(
            {u'name': item.get('name'), u'natIP': replace_resource_dict(item.get(u'nat_ip', {}), 'address'), u'type': item.get('type')}
        )

    def _response_from_item(self, item):
        return remove_nones_from_dict({u'name': item.get(u'name'), u'natIP': item.get(u'natIP'), u'type': item.get(u'type')})


class InstanceAliasiprangesArray(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = []

    def to_request(self):
        items = []
        for item in self.request:
            items.append(self._request_for_item(item))
        return items

    def from_response(self):
        items = []
        for item in self.request:
            items.append(self._response_from_item(item))
        return items

    def _request_for_item(self, item):
        return remove_nones_from_dict({u'ipCidrRange': item.get('ip_cidr_range'), u'subnetworkRangeName': item.get('subnetwork_range_name')})

    def _response_from_item(self, item):
        return remove_nones_from_dict({u'ipCidrRange': item.get(u'ipCidrRange'), u'subnetworkRangeName': item.get(u'subnetworkRangeName')})


class InstanceScheduling(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = {}

    def to_request(self):
        return remove_nones_from_dict(
            {
                u'automaticRestart': self.request.get('automatic_restart'),
                u'onHostMaintenance': self.request.get('on_host_maintenance'),
                u'preemptible': self.request.get('preemptible'),
            }
        )

    def from_response(self):
        return remove_nones_from_dict(
            {
                u'automaticRestart': self.request.get(u'automaticRestart'),
                u'onHostMaintenance': self.request.get(u'onHostMaintenance'),
                u'preemptible': self.request.get(u'preemptible'),
            }
        )


class InstanceServiceaccountsArray(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = []

    def to_request(self):
        items = []
        for item in self.request:
            items.append(self._request_for_item(item))
        return items

    def from_response(self):
        items = []
        for item in self.request:
            items.append(self._response_from_item(item))
        return items

    def _request_for_item(self, item):
        return remove_nones_from_dict({u'email': item.get('email'), u'scopes': item.get('scopes')})

    def _response_from_item(self, item):
        return remove_nones_from_dict({u'email': item.get(u'email'), u'scopes': item.get(u'scopes')})


class InstanceShieldedinstanceconfig(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = {}

    def to_request(self):
        return remove_nones_from_dict(
            {
                u'enableSecureBoot': self.request.get('enable_secure_boot'),
                u'enableVtpm': self.request.get('enable_vtpm'),
                u'enableIntegrityMonitoring': self.request.get('enable_integrity_monitoring'),
            }
        )

    def from_response(self):
        return remove_nones_from_dict(
            {
                u'enableSecureBoot': self.request.get(u'enableSecureBoot'),
                u'enableVtpm': self.request.get(u'enableVtpm'),
                u'enableIntegrityMonitoring': self.request.get(u'enableIntegrityMonitoring'),
            }
        )


class InstanceTags(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = {}

    def to_request(self):
        return remove_nones_from_dict({u'fingerprint': self.request.get('fingerprint'), u'items': self.request.get('items')})

    def from_response(self):
        return remove_nones_from_dict({u'fingerprint': self.request.get(u'fingerprint'), u'items': self.request.get(u'items')})


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Ansible module to manage Check Point Firewall (c) 2019
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import (absolute_import, division, print_function)

__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = """
---
module: cp_mgmt_threat_indicator
short_description: Manages threat-indicator objects on Check Point over Web Services API
description:
  - Manages threat-indicator objects on Check Point devices including creating, updating and removing objects.
  - All operations are performed over Web Services API.
version_added: "2.9"
author: "Or Soffer (@chkp-orso)"
options:
  name:
    description:
      - Object name.
    type: str
    required: True
  observables:
    description:
      - The indicator's observables.
    type: list
    suboptions:
      name:
        description:
          - Object name. Should be unique in the domain.
        type: str
      md5:
        description:
          - A valid MD5 sequence.
        type: str
      url:
        description:
          - A valid URL.
        type: str
      ip_address:
        description:
          - A valid IP-Address.
        type: str
      ip_address_first:
        description:
          - A valid IP-Address, the beginning of the range. If you configure this parameter with a value, you must also configure the value of the
            'ip-address-last' parameter.
        type: str
      ip_address_last:
        description:
          - A valid IP-Address, the end of the range. If you configure this parameter with a value, you must also configure the value of the
            'ip-address-first' parameter.
        type: str
      domain:
        description:
          - The name of a domain.
        type: str
      mail_to:
        description:
          - A valid E-Mail address, recipient filed.
        type: str
      mail_from:
        description:
          - A valid E-Mail address, sender field.
        type: str
      mail_cc:
        description:
          - A valid E-Mail address, cc field.
        type: str
      mail_reply_to:
        description:
          - A valid E-Mail address, reply-to field.
        type: str
      mail_subject:
        description:
          - Subject of E-Mail.
        type: str
      confidence:
        description:
          - The confidence level the indicator has that a real threat has been uncovered.
        type: str
        choices: ['low', 'medium', 'high', 'critical']
      product:
        description:
          - The software blade that processes the observable, AV - AntiVirus, AB - AntiBot.
        type: str
        choices: ['AV', 'AB']
      severity:
        description:
          - The severity level of the threat.
        type: str
        choices: ['low', 'medium', 'high', 'critical']
      comments:
        description:
          - Comments string.
        type: str
      ignore_warnings:
        description:
          - Apply changes ignoring warnings.
        type: bool
      ignore_errors:
        description:
          - Apply changes ignoring errors. You won't be able to publish such a changes. If ignore-warnings flag was omitted - warnings will also be ignored.
        type: bool
  observables_raw_data:
    description:
      - The contents of a file containing the indicator's observables.
    type: str
  action:
    description:
      - The indicator's action.
    type: str
    choices: ['Inactive', 'Ask', 'Prevent', 'Detect']
  profile_overrides:
    description:
      - Profiles in which to override the indicator's default action.
    type: list
    suboptions:
      action:
        description:
          - The indicator's action in this profile.
        type: str
        choices: ['Inactive', 'Ask', 'Prevent', 'Detect']
      profile:
        description:
          - The profile in which to override the indicator's action.
        type: str
  tags:
    description:
      - Collection of tag identifiers.
    type: list
  color:
    description:
      - Color of the object. Should be one of existing colors.
    type: str
    choices: ['aquamarine', 'black', 'blue', 'crete blue', 'burlywood', 'cyan', 'dark green', 'khaki', 'orchid', 'dark orange', 'dark sea green',
             'pink', 'turquoise', 'dark blue', 'firebrick', 'brown', 'forest green', 'gold', 'dark gold', 'gray', 'dark gray', 'light green', 'lemon chiffon',
             'coral', 'sea green', 'sky blue', 'magenta', 'purple', 'slate blue', 'violet red', 'navy blue', 'olive', 'orange', 'red', 'sienna', 'yellow']
  comments:
    description:
      - Comments string.
    type: str
  details_level:
    description:
      - The level of detail for some of the fields in the response can vary from showing only the UID value of the object to a fully detailed
        representation of the object.
    type: str
    choices: ['uid', 'standard', 'full']
  ignore_warnings:
    description:
      - Apply changes ignoring warnings.
    type: bool
  ignore_errors:
    description:
      - Apply changes ignoring errors. You won't be able to publish such a changes. If ignore-warnings flag was omitted - warnings will also be ignored.
    type: bool
extends_documentation_fragment: checkpoint_objects
"""

EXAMPLES = """
- name: add-threat-indicator
  cp_mgmt_threat_indicator:
    action: ask
    ignore_warnings: true
    name: My_Indicator
    observables:
    - confidence: medium
      mail_to: someone@somewhere.com
      name: My_Observable
      product: AV
      severity: low
    profile_overrides:
    - action: detect
      profile: My_Profile
    state: present

- name: set-threat-indicator
  cp_mgmt_threat_indicator:
    action: prevent
    ignore_warnings: true
    name: My_Indicator
    state: present

- name: delete-threat-indicator
  cp_mgmt_threat_indicator:
    name: My_Indicator
    state: absent
"""

RETURN = """
cp_mgmt_threat_indicator:
  description: The checkpoint object created or updated.
  returned: always, except when deleting the object.
  type: dict
"""

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.network.checkpoint.checkpoint import checkpoint_argument_spec_for_objects, api_call


def main():
    argument_spec = dict(
        name=dict(type='str', required=True),
        observables=dict(type='list', options=dict(
            name=dict(type='str'),
            md5=dict(type='str'),
            url=dict(type='str'),
            ip_address=dict(type='str'),
            ip_address_first=dict(type='str'),
            ip_address_last=dict(type='str'),
            domain=dict(type='str'),
            mail_to=dict(type='str'),
            mail_from=dict(type='str'),
            mail_cc=dict(type='str'),
            mail_reply_to=dict(type='str'),
            mail_subject=dict(type='str'),
            confidence=dict(type='str', choices=['low', 'medium', 'high', 'critical']),
            product=dict(type='str', choices=['AV', 'AB']),
            severity=dict(type='str', choices=['low', 'medium', 'high', 'critical']),
            comments=dict(type='str'),
            ignore_warnings=dict(type='bool'),
            ignore_errors=dict(type='bool')
        )),
        observables_raw_data=dict(type='str'),
        action=dict(type='str', choices=['Inactive', 'Ask', 'Prevent', 'Detect']),
        profile_overrides=dict(type='list', options=dict(
            action=dict(type='str', choices=['Inactive', 'Ask', 'Prevent', 'Detect']),
            profile=dict(type='str')
        )),
        tags=dict(type='list'),
        color=dict(type='str', choices=['aquamarine', 'black', 'blue', 'crete blue', 'burlywood', 'cyan', 'dark green',
                                        'khaki', 'orchid', 'dark orange', 'dark sea green', 'pink', 'turquoise', 'dark blue', 'firebrick', 'brown',
                                        'forest green', 'gold', 'dark gold', 'gray', 'dark gray', 'light green', 'lemon chiffon', 'coral', 'sea green',
                                        'sky blue', 'magenta', 'purple', 'slate blue', 'violet red', 'navy blue', 'olive', 'orange', 'red', 'sienna',
                                        'yellow']),
        comments=dict(type='str'),
        details_level=dict(type='str', choices=['uid', 'standard', 'full']),
        ignore_warnings=dict(type='bool'),
        ignore_errors=dict(type='bool')
    )
    argument_spec.update(checkpoint_argument_spec_for_objects)

    module = AnsibleModule(argument_spec=argument_spec, supports_check_mode=True)
    api_call_object = 'threat-indicator'

    result = api_call(module, api_call_object)
    module.exit_json(**result)


if __name__ == '__main__':
    main()



# (c) 2012-2014, Michael DeHaan <michael.dehaan@gmail.com>
# (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

DOCUMENTATION = '''
    callback: default
    type: stdout
    short_description: default Ansible screen output
    version_added: historical
    description:
        - This is the default output callback for ansible-playbook.
    extends_documentation_fragment:
      - default_callback
    requirements:
      - set as stdout in configuration
    options:
      check_mode_markers:
        name: Show markers when running in check mode
        description:
        - "Toggle to control displaying markers when running in check mode. The markers are C(DRY RUN)
        at the beggining and ending of playbook execution (when calling C(ansible-playbook --check))
        and C(CHECK MODE) as a suffix at every play and task that is run in check mode."
        type: bool
        default: no
        version_added: 2.9
        env:
          - name: ANSIBLE_CHECK_MODE_MARKERS
        ini:
          - key: check_mode_markers
            section: defaults
'''

# NOTE: check_mode_markers functionality is also implemented in the following derived plugins:
#       debug.py, yaml.py, dense.py. Maybe their documentation needs updating, too.


from ansible import constants as C
from ansible import context
from ansible.playbook.task_include import TaskInclude
from ansible.plugins.callback import CallbackBase
from ansible.utils.color import colorize, hostcolor


# These values use ansible.constants for historical reasons, mostly to allow
# unmodified derivative plugins to work. However, newer options added to the
# plugin are not also added to ansible.constants, so authors of derivative
# callback plugins will eventually need to add a reference to the common docs
# fragment for the 'default' callback plugin

# these are used to provide backwards compat with old plugins that subclass from default
# but still don't use the new config system and/or fail to document the options
# TODO: Change the default of check_mode_markers to True in a future release (2.13)
COMPAT_OPTIONS = (('display_skipped_hosts', C.DISPLAY_SKIPPED_HOSTS),
                  ('display_ok_hosts', True),
                  ('show_custom_stats', C.SHOW_CUSTOM_STATS),
                  ('display_failed_stderr', False),
                  ('check_mode_markers', False),)


class CallbackModule(CallbackBase):

    '''
    This is the default callback interface, which simply prints messages
    to stdout when new callback events are received.
    '''

    CALLBACK_VERSION = 2.0
    CALLBACK_TYPE = 'stdout'
    CALLBACK_NAME = 'default'

    def __init__(self):

        self._play = None
        self._last_task_banner = None
        self._last_task_name = None
        self._task_type_cache = {}
        super(CallbackModule, self).__init__()

    def set_options(self, task_keys=None, var_options=None, direct=None):

        super(CallbackModule, self).set_options(task_keys=task_keys, var_options=var_options, direct=direct)

        # for backwards compat with plugins subclassing default, fallback to constants
        for option, constant in COMPAT_OPTIONS:
            try:
                value = self.get_option(option)
            except (AttributeError, KeyError):
                value = constant
            setattr(self, option, value)

    def v2_runner_on_failed(self, result, ignore_errors=False):

        delegated_vars = result._result.get('_ansible_delegated_vars', None)
        self._clean_results(result._result, result._task.action)

        if self._last_task_banner != result._task._uuid:
            self._print_task_banner(result._task)

        self._handle_exception(result._result, use_stderr=self.display_failed_stderr)
        self._handle_warnings(result._result)

        if result._task.loop and 'results' in result._result:
            self._process_items(result)

        else:
            if delegated_vars:
                self._display.display("fatal: [%s -> %s]: FAILED! => %s" % (result._host.get_name(), delegated_vars['ansible_host'],
                                                                            self._dump_results(result._result)),
                                      color=C.COLOR_ERROR, stderr=self.display_failed_stderr)
            else:
                self._display.display("fatal: [%s]: FAILED! => %s" % (result._host.get_name(), self._dump_results(result._result)),
                                      color=C.COLOR_ERROR, stderr=self.display_failed_stderr)

        if ignore_errors:
            self._display.display("...ignoring", color=C.COLOR_SKIP)

    def v2_runner_on_ok(self, result):

        delegated_vars = result._result.get('_ansible_delegated_vars', None)

        if isinstance(result._task, TaskInclude):
            return
        elif result._result.get('changed', False):
            if self._last_task_banner != result._task._uuid:
                self._print_task_banner(result._task)

            if delegated_vars:
                msg = "changed: [%s -> %s]" % (result._host.get_name(), delegated_vars['ansible_host'])
            else:
                msg = "changed: [%s]" % result._host.get_name()
            color = C.COLOR_CHANGED
        else:
            if not self.display_ok_hosts:
                return

            if self._last_task_banner != result._task._uuid:
                self._print_task_banner(result._task)

            if delegated_vars:
                msg = "ok: [%s -> %s]" % (result._host.get_name(), delegated_vars['ansible_host'])
            else:
                msg = "ok: [%s]" % result._host.get_name()
            color = C.COLOR_OK

        self._handle_warnings(result._result)

        if result._task.loop and 'results' in result._result:
            self._process_items(result)
        else:
            self._clean_results(result._result, result._task.action)

            if self._run_is_verbose(result):
                msg += " => %s" % (self._dump_results(result._result),)
            self._display.display(msg, color=color)

    def v2_runner_on_skipped(self, result):

        if self.display_skipped_hosts:

            self._clean_results(result._result, result._task.action)

            if self._last_task_banner != result._task._uuid:
                self._print_task_banner(result._task)

            if result._task.loop and 'results' in result._result:
                self._process_items(result)
            else:
                msg = "skipping: [%s]" % result._host.get_name()
                if self._run_is_verbose(result):
                    msg += " => %s" % self._dump_results(result._result)
                self._display.display(msg, color=C.COLOR_SKIP)

    def v2_runner_on_unreachable(self, result):
        if self._last_task_banner != result._task._uuid:
            self._print_task_banner(result._task)

        delegated_vars = result._result.get('_ansible_delegated_vars', None)
        if delegated_vars:
            msg = "fatal: [%s -> %s]: UNREACHABLE! => %s" % (result._host.get_name(), delegated_vars['ansible_host'], self._dump_results(result._result))
        else:
            msg = "fatal: [%s]: UNREACHABLE! => %s" % (result._host.get_name(), self._dump_results(result._result))
        self._display.display(msg, color=C.COLOR_UNREACHABLE, stderr=self.display_failed_stderr)

    def v2_playbook_on_no_hosts_matched(self):
        self._display.display("skipping: no hosts matched", color=C.COLOR_SKIP)

    def v2_playbook_on_no_hosts_remaining(self):
        self._display.banner("NO MORE HOSTS LEFT")

    def v2_playbook_on_task_start(self, task, is_conditional):
        self._task_start(task, prefix='TASK')

    def _task_start(self, task, prefix=None):
        # Cache output prefix for task if provided
        # This is needed to properly display 'RUNNING HANDLER' and similar
        # when hiding skipped/ok task results
        if prefix is not None:
            self._task_type_cache[task._uuid] = prefix

        # Preserve task name, as all vars may not be available for templating
        # when we need it later
        if self._play.strategy == 'free':
            # Explicitly set to None for strategy 'free' to account for any cached
            # task title from a previous non-free play
            self._last_task_name = None
        else:
            self._last_task_name = task.get_name().strip()

            # Display the task banner immediately if we're not doing any filtering based on task result
            if self.display_skipped_hosts and self.display_ok_hosts:
                self._print_task_banner(task)

    def _print_task_banner(self, task):
        # args can be specified as no_log in several places: in the task or in
        # the argument spec.  We can check whether the task is no_log but the
        # argument spec can't be because that is only run on the target
        # machine and we haven't run it thereyet at this time.
        #
        # So we give people a config option to affect display of the args so
        # that they can secure this if they feel that their stdout is insecure
        # (shoulder surfing, logging stdout straight to a file, etc).
        args = ''
        if not task.no_log and C.DISPLAY_ARGS_TO_STDOUT:
            args = u', '.join(u'%s=%s' % a for a in task.args.items())
            args = u' %s' % args

        prefix = self._task_type_cache.get(task._uuid, 'TASK')

        # Use cached task name
        task_name = self._last_task_name
        if task_name is None:
            task_name = task.get_name().strip()

        if task.check_mode and self.check_mode_markers:
            checkmsg = " [CHECK MODE]"
        else:
            checkmsg = ""
        self._display.banner(u"%s [%s%s]%s" % (prefix, task_name, args, checkmsg))
        if self._display.verbosity >= 2:
            path = task.get_path()
            if path:
                self._display.display(u"task path: %s" % path, color=C.COLOR_DEBUG)

        self._last_task_banner = task._uuid

    def v2_playbook_on_cleanup_task_start(self, task):
        self._task_start(task, prefix='CLEANUP TASK')

    def v2_playbook_on_handler_task_start(self, task):
        self._task_start(task, prefix='RUNNING HANDLER')

    def v2_runner_on_start(self, host, task):
        if self.get_option('show_per_host_start'):
            self._display.display(" [started %s on %s]" % (task, host), color=C.COLOR_OK)

    def v2_playbook_on_play_start(self, play):
        name = play.get_name().strip()
        if play.check_mode and self.check_mode_markers:
            checkmsg = " [CHECK MODE]"
        else:
            checkmsg = ""
        if not name:
            msg = u"PLAY%s" % checkmsg
        else:
            msg = u"PLAY [%s]%s" % (name, checkmsg)

        self._play = play

        self._display.banner(msg)

    def v2_on_file_diff(self, result):
        if result._task.loop and 'results' in result._result:
            for res in result._result['results']:
                if 'diff' in res and res['diff'] and res.get('changed', False):
                    diff = self._get_diff(res['diff'])
                    if diff:
                        if self._last_task_banner != result._task._uuid:
                            self._print_task_banner(result._task)
                        self._display.display(diff)
        elif 'diff' in result._result and result._result['diff'] and result._result.get('changed', False):
            diff = self._get_diff(result._result['diff'])
            if diff:
                if self._last_task_banner != result._task._uuid:
                    self._print_task_banner(result._task)
                self._display.display(diff)

    def v2_runner_item_on_ok(self, result):

        delegated_vars = result._result.get('_ansible_delegated_vars', None)
        if isinstance(result._task, TaskInclude):
            return
        elif result._result.get('changed', False):
            if self._last_task_banner != result._task._uuid:
                self._print_task_banner(result._task)

            msg = 'changed'
            color = C.COLOR_CHANGED
        else:
            if not self.display_ok_hosts:
                return

            if self._last_task_banner != result._task._uuid:
                self._print_task_banner(result._task)

            msg = 'ok'
            color = C.COLOR_OK

        if delegated_vars:
            msg += ": [%s -> %s]" % (result._host.get_name(), delegated_vars['ansible_host'])
        else:
            msg += ": [%s]" % result._host.get_name()

        msg += " => (item=%s)" % (self._get_item_label(result._result),)

        self._clean_results(result._result, result._task.action)
        if self._run_is_verbose(result):
            msg += " => %s" % self._dump_results(result._result)
        self._display.display(msg, color=color)

    def v2_runner_item_on_failed(self, result):
        if self._last_task_banner != result._task._uuid:
            self._print_task_banner(result._task)

        delegated_vars = result._result.get('_ansible_delegated_vars', None)
        self._clean_results(result._result, result._task.action)
        self._handle_exception(result._result)

        msg = "failed: "
        if delegated_vars:
            msg += "[%s -> %s]" % (result._host.get_name(), delegated_vars['ansible_host'])
        else:
            msg += "[%s]" % (result._host.get_name())

        self._handle_warnings(result._result)
        self._display.display(msg + " (item=%s) => %s" % (self._get_item_label(result._result), self._dump_results(result._result)), color=C.COLOR_ERROR)

    def v2_runner_item_on_skipped(self, result):
        if self.display_skipped_hosts:
            if self._last_task_banner != result._task._uuid:
                self._print_task_banner(result._task)

            self._clean_results(result._result, result._task.action)
            msg = "skipping: [%s] => (item=%s) " % (result._host.get_name(), self._get_item_label(result._result))
            if self._run_is_verbose(result):
                msg += " => %s" % self._dump_results(result._result)
            self._display.display(msg, color=C.COLOR_SKIP)

    def v2_playbook_on_include(self, included_file):
        msg = 'included: %s for %s' % (included_file._filename, ", ".join([h.name for h in included_file._hosts]))
        if 'item' in included_file._args:
            msg += " => (item=%s)" % (self._get_item_label(included_file._args),)
        self._display.display(msg, color=C.COLOR_SKIP)

    def v2_playbook_on_stats(self, stats):
        self._display.banner("PLAY RECAP")

        hosts = sorted(stats.processed.keys())
        for h in hosts:
            t = stats.summarize(h)

            self._display.display(
                u"%s : %s %s %s %s %s %s %s" % (
                    hostcolor(h, t),
                    colorize(u'ok', t['ok'], C.COLOR_OK),
                    colorize(u'changed', t['changed'], C.COLOR_CHANGED),
                    colorize(u'unreachable', t['unreachable'], C.COLOR_UNREACHABLE),
                    colorize(u'failed', t['failures'], C.COLOR_ERROR),
                    colorize(u'skipped', t['skipped'], C.COLOR_SKIP),
                    colorize(u'rescued', t['rescued'], C.COLOR_OK),
                    colorize(u'ignored', t['ignored'], C.COLOR_WARN),
                ),
                screen_only=True
            )

            self._display.display(
                u"%s : %s %s %s %s %s %s %s" % (
                    hostcolor(h, t, False),
                    colorize(u'ok', t['ok'], None),
                    colorize(u'changed', t['changed'], None),
                    colorize(u'unreachable', t['unreachable'], None),
                    colorize(u'failed', t['failures'], None),
                    colorize(u'skipped', t['skipped'], None),
                    colorize(u'rescued', t['rescued'], None),
                    colorize(u'ignored', t['ignored'], None),
                ),
                log_only=True
            )

        self._display.display("", screen_only=True)

        # print custom stats if required
        if stats.custom and self.show_custom_stats:
            self._display.banner("CUSTOM STATS: ")
            # per host
            # TODO: come up with 'pretty format'
            for k in sorted(stats.custom.keys()):
                if k == '_run':
                    continue
                self._display.display('\t%s: %s' % (k, self._dump_results(stats.custom[k], indent=1).replace('\n', '')))

            # print per run custom stats
            if '_run' in stats.custom:
                self._display.display("", screen_only=True)
                self._display.display('\tRUN: %s' % self._dump_results(stats.custom['_run'], indent=1).replace('\n', ''))
            self._display.display("", screen_only=True)

        if context.CLIARGS['check'] and self.check_mode_markers:
            self._display.banner("DRY RUN")

    def v2_playbook_on_start(self, playbook):
        if self._display.verbosity > 1:
            from os.path import basename
            self._display.banner("PLAYBOOK: %s" % basename(playbook._file_name))

        # show CLI arguments
        if self._display.verbosity > 3:
            if context.CLIARGS.get('args'):
                self._display.display('Positional arguments: %s' % ' '.join(context.CLIARGS['args']),
                                      color=C.COLOR_VERBOSE, screen_only=True)

            for argument in (a for a in context.CLIARGS if a != 'args'):
                val = context.CLIARGS[argument]
                if val:
                    self._display.display('%s: %s' % (argument, val), color=C.COLOR_VERBOSE, screen_only=True)

        if context.CLIARGS['check'] and self.check_mode_markers:
            self._display.banner("DRY RUN")

    def v2_runner_retry(self, result):
        task_name = result.task_name or result._task
        msg = "FAILED - RETRYING: %s (%d retries left)." % (task_name, result._result['retries'] - result._result['attempts'])
        if self._run_is_verbose(result, verbosity=2):
            msg += "Result was: %s" % self._dump_results(result._result)
        self._display.display(msg, color=C.COLOR_DEBUG)

    def v2_playbook_on_notify(self, handler, host):
        if self._display.verbosity > 1:
            self._display.display("NOTIFIED HANDLER %s for %s" % (handler.get_name(), host), color=C.COLOR_VERBOSE, screen_only=True)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright: (c) 2019, F5 Networks Inc.
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'certified'}

DOCUMENTATION = r'''
---
module: bigip_device_certificate
short_description: Manage self-signed device certificates
description:
  - Module used to create and/or renew self-signed device certificates for BIG-IP.
version_added: 2.9
options:
  days_valid:
    description:
      - Specifies the interval for which the self-signed certificate is valid.
      - "The maximum value is 25 years: C(9125) days"
    type: int
    required: True
  cert_name:
    description:
      - Specifies the full name of the certificate file.
      - If the name is not default C(server.crt), the module will configure C(httpd) to use them
        prior to restarting the C(httpd) daemon.
    type: str
    default: server.crt
  key_name:
    description:
      - Specifies the full name of the key file.
      - If the name is not default C(server.key), the module will configure C(httpd) to use them
        prior to restarting the C(httpd) daemon.
    type: str
    default: server.key
  key_size:
    description:
      - Specifies the desired key size in bits.
      - Mandatory option when generating a new certificate.
    type: int
    choices:
      - 512
      - 1024
      - 2048
      - 4096
    default: 2048
  issuer:
    description:
      - Certificate properties, required when generating new certificates.
    suboptions:
      country:
        description:
         - Specifies the Country name attribute for the certificate.
        type: str
      state:
        description:
          - Specifies the State or Province attribute for the certificate.
        type: str
      locality:
        description:
          - Specifies the city or town name for the certificate.
        type: str
      organization:
        description:
          - Specifies the Organization attribute for the certificate.
        type: str
      division:
        description:
          - Specifies the department name attribute for the certificate.
        type: str
      common_name:
        description:
          - Specifies Common Name attribute for the certificate.
        type: str
      email:
        description:
          - "Specifies the domain administrator's email address."
        type: str
    type: dict
  add_to_trusted:
    description:
      - Specified if the certificate should be added to the trusted client and server certificate files.
    type: bool
    default: no
  new_cert:
    description:
      - Specified if the module should generate new certificate.
      - When C(yes) the device certificate and key will be replaced
    type: bool
    default: no
  force:
    description:
      - When C(yes), will update or overwrite the existing certificate when it is not expired device.
      - When C(no), the certificate will only be updated/overwritten if expired.
      - Generally should be C(yes) only in cases where you need to update certificate that is about to expire.
      - This option is also needed when generating new certificate to replace non expired one.
    type: bool
    default: no
  transport:
    description:
      - Configures the transport connection to use when connecting to the
        remote device.
      - This module currently supports only connectivity to the device over cli (ssh).
    required: True
    choices:
        - cli
    default: cli
extends_documentation_fragment: f5
author:
  - Wojciech Wypior (@wojtek0806)
'''

EXAMPLES = r'''
- name: Update expired certificate
  bigip_device_certificate:
    days_valid: 365
    provider:
      password: secret
      server: lb.mydomain.com
      user: admin
      transport: cli
      server_port: 22
  delegate_to: localhost

- name: Update expired certificate non-default names
  bigip_device_certificate:
    days_valid: 60
    cert_name: custom.crt
    key_name: custom.key
    provider:
      password: secret
      server: lb.mydomain.com
      user: admin
      transport: cli
      server_port: 22
  delegate_to: localhost

- name: Force update not expired certificate
  bigip_device_certificate:
    days_valid: 365
    force: yes
    provider:
      password: secret
      server: lb.mydomain.com
      user: admin
      transport: cli
      server_port: 22
  delegate_to: localhost

- name: Create a new certificate to replace expired certificate
  bigip_device_certificate:
    days_valid: 365
    new_cert: yes
    issuer:
      country: US
      state: WA
      common_name: foobar.foo.local
    provider:
      password: secret
      server: lb.mydomain.com
      user: admin
      transport: cli
      server_port: 22
  delegate_to: localhost

- name: Force create a new custom named certificate to replace not expired certificate
  bigip_device_certificate:
    days_valid: 365
    cert_name: custom.crt
    key_name: custom.key
    new_cert: yes
    force: yes
    issuer:
      country: US
      state: WA
      common_name: foobar.foo.local
    key_size: 2048
    provider:
      password: secret
      server: lb.mydomain.com
      user: admin
      transport: cli
      server_port: 22
  delegate_to: localhost
'''

RETURN = r'''
days_valid:
  description: The interval for which the self-signed certificate is valid.
  returned: changed
  type: int
  sample: 365
issuer:
  description: Specifies certificate properties.
  type: complex
  returned: changed
  contains:
    country:
      description: The Country name attribute of the certificate.
      returned: changed
      type: str
      sample: US
    state:
      description: The State or Province attribute of the certificate.
      returned: changed
      type: str
      sample: WA
    locality:
      description: The city or town name attribute of the certificate.
      returned: changed
      type: str
      sample: Seattle
    organization:
      description: The Organization attribute of the certificate.
      returned: changed
      type: str
      sample: F5
    division:
      description: The department name attribute of the certificate.
      returned: changed
      type: str
      sample: IT
    common_name:
      description: The Common Name attribute of the certificate.
      returned: changed
      type: str
      sample: foo.bar.local
    email:
      description: "The domain administrator's email address."
      returned: changed
      type: str
      sample: admin@foo.bar.local
cert_name:
  description: The full name of the certificate file.
  returned: changed
  type: str
  sample: common.crt
key_name:
  description: The full name of the key file.
  returned: changed
  type: str
  sample: common.key
key_size:
  description: The desired key size in bits.
  returned: changed
  type: int
  sample: 2048
'''
import os
import ssl
from datetime import datetime

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.connection import exec_command


try:
    from library.module_utils.network.f5.common import F5ModuleError
    from library.module_utils.network.f5.common import AnsibleF5Parameters
    from library.module_utils.network.f5.common import f5_argument_spec
    from library.module_utils.network.f5.common import is_cli
except ImportError:
    from ansible.module_utils.network.f5.common import F5ModuleError
    from ansible.module_utils.network.f5.common import AnsibleF5Parameters
    from ansible.module_utils.network.f5.common import f5_argument_spec
    from ansible.module_utils.network.f5.common import is_cli


class Parameters(AnsibleF5Parameters):
    returnables = [
        'days_valid',
        'issuer',
        'key_size',
        'cert_name',
        'key_name',
    ]

    updatables = [
        'days_valid',
    ]


class ApiParameters(Parameters):
    pass


class ModuleParameters(Parameters):
    issuer_map = {
        'country': 'C',
        'state': 'ST',
        'locality': 'L',
        'organization': 'O',
        'division': 'OU',
        'common_name': 'CN',
        'email': 'emailAddress'
    }

    @property
    def issuer(self):
        if self._values['issuer'] is None:
            return None
        filtered = dict((self.issuer_map[k], v) for k, v in self._values['issuer'].items() if v is not None)
        to_parse = ['{0}={1}'.format(k, v) for k, v in filtered.items()]
        result = '/' + '/'.join(to_parse) + '/'
        return result

    @property
    def days_valid(self):
        if 1 <= self._values['days_valid'] <= 9125:
            return self._values['days_valid']
        raise F5ModuleError(
            "Valid 'days_valid' must be in range 1 - 9125 days."
        )


class Changes(Parameters):
    def to_return(self):
        result = {}
        try:
            for returnable in self.returnables:
                result[returnable] = getattr(self, returnable)
            result = self._filter_params(result)
        except Exception:
            pass
        return result


class UsableChanges(Changes):
    pass


class ReportableChanges(Changes):
    issuer_map = {
        'C': 'country',
        'ST': 'state',
        'L': 'locality',
        'O': 'organization',
        'OU': 'division',
        'CN': 'common_name',
        'emailAddress': 'email'
    }

    @property
    def issuer(self):
        if self._values['issuer'] is None:
            return None
        to_dict = [tuple(item.split('=')) for item in self._values['issuer'].strip('/').split('/')]
        result = dict((self.issuer_map[k], v) for k, v in to_dict)
        return result


class Difference(object):
    def __init__(self, want, have=None):
        self.want = want
        self.have = have

    def compare(self, param):
        try:
            result = getattr(self, param)
            return result
        except AttributeError:
            return self.__default(param)

    def __default(self, param):
        attr1 = getattr(self.want, param)
        try:
            attr2 = getattr(self.have, param)
            if attr1 != attr2:
                return attr1
        except AttributeError:
            return attr1


class ModuleManager(object):
    def __init__(self, *args, **kwargs):
        self.module = kwargs.get('module', None)
        self.want = ModuleParameters(params=self.module.params)
        self.have = ApiParameters()
        self.changes = UsableChanges()

    def _set_changed_options(self):
        changed = {}
        for key in Parameters.returnables:
            if getattr(self.want, key) is not None:
                changed[key] = getattr(self.want, key)
        if changed:
            self.changes = UsableChanges(params=changed)

    def _update_changed_options(self):
        diff = Difference(self.want, self.have)
        updatables = Parameters.updatables
        changed = dict()
        for k in updatables:
            change = diff.compare(k)
            if change is None:
                continue
            else:
                if isinstance(change, dict):
                    changed.update(change)
                else:
                    changed[k] = change
        if changed:
            self.changes = UsableChanges(params=changed)
            return True
        return False

    def exec_module(self):
        if not is_cli(self.module):
            raise F5ModuleError('Module can only be run via SSH, set the transport property to CLI')

        result = dict()

        changed = self.present()

        reportable = ReportableChanges(params=self.changes.to_return())
        changes = reportable.to_return()
        result.update(**changes)
        result.update(dict(changed=changed))
        return result

    def present(self):
        if self.expired():
            if self.want.new_cert:
                self.create()
                return True
            self.update()
            return True
        if self.want.force and self.want.new_cert:
            self.create()
            return True
        if self.want.force:
            self.update()
            return True
        return False

    def create(self):
        self._set_changed_options()
        if self.module.check_mode:
            return True
        self.generate_new()
        return True

    def update(self):
        self._update_changed_options()
        if self.module.check_mode:
            return True
        self.update_certificate()
        if self.want.cert_name != 'server.crt' or self.want.key_name != 'server.key':
            self.configure_new_cert()
        self.restart_daemon()
        if self.want.add_to_trusted:
            self.copy_files_to_trusted()
        return True

    def expired(self):
        self.have = self.read_current_certificate()
        current_epoch = int(datetime.now().timestamp())
        if current_epoch > self.have.epoch:
            return True
        return False

    def generate_new(self):
        self.generate_cert_key()
        if self.want.cert_name != 'server.crt' or self.want.key_name != 'server.key':
            self.configure_new_cert()
        self.restart_daemon()
        if self.want.add_to_trusted:
            self.copy_files_to_trusted()
        return True

    def generate_cert_key(self):
        cmd = 'openssl req -x509 -nodes -days {3} -newkey rsa:{4} -keyout {0}/ssl.key/{2} ' \
              '-out {0}/ssl.crt/{1} -subj "{5}"'.format('/config/httpd/conf', self.want.cert_name, self.want.key_name,
                                                        self.want.days_valid, self.want.key_size, self.want.issuer)
        rc, out, err = exec_command(self.module, cmd)
        if rc != 0:
            raise F5ModuleError(err)

    def create_csr(self):
        cmd = 'openssl x509 -x509toreq -in {0}/ssl.crt/{1} -out {0}/ssl.csr/{3}.csr -signkey {0}/ssl.key/{2}'.format(
            '/config/httpd/conf', self.want.cert_name, self.want.key_name, os.path.splitext(self.want.cert_name)[0]
        )

        rc, out, err = exec_command(self.module, cmd)
        if rc != 0:
            raise F5ModuleError(err)

    def update_certificate(self):
        self.create_csr()
        cmd = 'openssl x509 -req -in {0}/ssl.csr/{3}.csr -signkey {0}/ssl.key/{2} -days {4} -out {0}/ssl.crt/{1}'.\
            format('/config/httpd/conf', self.want.cert_name, self.want.key_name,
                   os.path.splitext(self.want.cert_name)[0], self.want.days_valid)
        rc, out, err = exec_command(self.module, cmd)
        if rc != 0:
            raise F5ModuleError(err)

    def configure_new_cert(self):
        cmd1 = 'tmsh modify sys httpd ssl-certkeyfile /config/httpd/conf/ssl.key/{1}' \
               'ssl-certfile /config/httpd/conf/ssl.crt/{0}'.format(self.want.cert_name, self.want.key_name)

        cmd2 = 'tmsh save /sys config partitions all'

        rc, out, err = exec_command(self.module, cmd1)
        if rc != 0:
            raise F5ModuleError(err)

        rc, out, err = exec_command(self.module, cmd2)
        if rc != 0:
            raise F5ModuleError(err)

    def restart_daemon(self):
        cmd = 'tmsh restart /sys service httpd'
        rc, out, err = exec_command(self.module, cmd)
        if rc != 0:
            raise F5ModuleError(err)

    def copy_files_to_trusted(self):
        cmd1 = 'cat /config/httpd/conf/ssl.crt/{0} >> /config/big3d/client.crt'.format(self.want.cert_name)
        cmd2 = 'cat /config/httpd/conf/ssl.crt/{0} >> /config/gtm/server.crt'.format(self.want.cert_name)
        rc, out, err = exec_command(self.module, cmd1)
        if rc != 0:
            raise F5ModuleError(err)
        rc, out, err = exec_command(self.module, cmd2)
        if rc != 0:
            raise F5ModuleError(err)

    def read_current_certificate(self):
        result = dict()
        command = 'openssl x509 -in /config/httpd/conf/ssl.crt/{0} -dates -issuer -noout'.format(self.want.cert_name)
        rc, out, err = exec_command(self.module, command)
        if rc == 0:
            result['epoch'] = self._parse_cert_date(out)
        return ApiParameters(params=result)

    def _parse_cert_date(self, to_parse):
        c_time = to_parse.split('\n')[1].split('=')[1]
        result = ssl.cert_time_to_seconds(c_time)
        return result


class ArgumentSpec(object):
    def __init__(self):
        self.supports_check_mode = True
        argument_spec = dict(
            key_size=dict(
                type='int',
                choices=[512, 1024, 2048, 4096],
                default=2048
            ),
            cert_name=dict(
                default='server.crt'
            ),
            key_name=dict(
                default='server.key'
            ),
            days_valid=dict(
                type='int',
                required=True
            ),
            issuer=dict(
                type='dict',
                options=dict(
                    country=dict(),
                    state=dict(),
                    locality=dict(),
                    organization=dict(),
                    division=dict(),
                    common_name=dict(),
                    email=dict(),
                ),
                required_one_of=[
                    ['country', 'state', 'locality', 'organization', 'division', 'common_name', 'email']
                ]
            ),

            add_to_trusted=dict(
                type='bool',
                default='no'
            ),
            new_cert=dict(
                type='bool',
                default='no'
            ),
            force=dict(
                type='bool',
                default='no'
            ),
            transport=dict(
                type='str',
                default='cli',
                choices=['cli']
            ),
        )
        self.argument_spec = {}
        self.argument_spec.update(f5_argument_spec)
        self.argument_spec.update(argument_spec)
        self.required_if = [
            ['new_cert', 'yes', ['days_valid', 'issuer', 'key_size']]
        ]


def main():
    spec = ArgumentSpec()

    module = AnsibleModule(
        argument_spec=spec.argument_spec,
        supports_check_mode=spec.supports_check_mode,
        required_if=spec.required_if,
    )

    try:
        mm = ModuleManager(module=module)
        results = mm.exec_module()
        module.exit_json(**results)
    except F5ModuleError as ex:
        module.fail_json(msg=str(ex))


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2018, Dag Wieers (@dagwieers) <dag@wieers.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = r'''
---
module: mso_schema
short_description: Manage schemas
description:
- Manage schemas on Cisco ACI Multi-Site.
author:
- Dag Wieers (@dagwieers)
version_added: '2.8'
options:
  schema:
    description:
    - The name of the schema.
    type: str
    required: yes
    aliases: [ name ]
  templates:
    description:
    - A list of templates for this schema.
    type: list
  sites:
    description:
    - A list of sites mapped to templates in this schema.
    type: list
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
notes:
- Due to restrictions of the MSO REST API this module cannot create empty schemas (i.e. schemas without templates).
  Use the M(mso_schema_template) to automatically create schemas with templates.
seealso:
- module: mso_schema_site
- module: mso_schema_template
extends_documentation_fragment: mso
'''

EXAMPLES = r'''
- name: Add a new schema
  mso_schema:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema 1
    state: present
    templates:
    - name: Template1
      displayName: Template 1
      tenantId: north_europe
      anps:
        <...>
    - name: Template2
      displayName: Template 2
      tenantId: nort_europe
      anps:
        <...>
  delegate_to: localhost

- name: Remove schemas
  mso_schema:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema 1
    state: absent
  delegate_to: localhost

- name: Query a schema
  mso_schema:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema 1
    state: query
  delegate_to: localhost
  register: query_result

- name: Query all schemas
  mso_schema:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    state: query
  delegate_to: localhost
  register: query_result
'''

RETURN = r'''
'''

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.network.aci.mso import MSOModule, mso_argument_spec, issubset


def main():
    argument_spec = mso_argument_spec()
    argument_spec.update(
        schema=dict(type='str', aliases=['name']),
        templates=dict(type='list'),
        sites=dict(type='list'),
        # messages=dict(type='dict'),
        # associations=dict(type='list'),
        # health_faults=dict(type='list'),
        # references=dict(type='dict'),
        # policy_states=dict(type='list'),
        state=dict(type='str', default='present', choices=['absent', 'present', 'query']),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ['state', 'absent', ['schema']],
            ['state', 'present', ['schema', 'templates']],
        ],
    )

    schema = module.params['schema']
    templates = module.params['templates']
    sites = module.params['sites']
    state = module.params['state']

    mso = MSOModule(module)

    schema_id = None
    path = 'schemas'

    # Query for existing object(s)
    if schema:
        mso.existing = mso.get_obj(path, displayName=schema)
        if mso.existing:
            schema_id = mso.existing['id']
            path = 'schemas/{id}'.format(id=schema_id)
    else:
        mso.existing = mso.query_objs(path)

    if state == 'query':
        pass

    elif state == 'absent':
        mso.previous = mso.existing
        if mso.existing:
            if module.check_mode:
                mso.existing = {}
            else:
                mso.existing = mso.request(path, method='DELETE')

    elif state == 'present':
        mso.previous = mso.existing

        payload = dict(
            id=schema_id,
            displayName=schema,
            templates=templates,
            sites=sites,
        )

        mso.sanitize(payload, collate=True)

        if mso.existing:
            if not issubset(mso.sent, mso.existing):
                if module.check_mode:
                    mso.existing = mso.proposed
                else:
                    mso.existing = mso.request(path, method='PUT', data=mso.sent)
        else:
            if module.check_mode:
                mso.existing = mso.proposed
            else:
                mso.existing = mso.request(path, method='POST', data=mso.sent)

    mso.exit_json()


if __name__ == "__main__":
    main()
